from gpglib.content_parsers.base import Parser
from gpglib.content_parsers.data import (
    CompressedParser,
    LiteralParser,
    ModificationDetectionCodePacketParser,
    SymEncryptedParser,
    SymmetricallyEncryptedIntegrityProtectedDataPacketParser,
    UserIdParser,
)
from gpglib.content_parsers.keys import (
    PublicKeyParser,
    PublicSubKeyParser,
    SecretKeyParser,
    SecretSubKeyParser,
    SignatureParser,
)
from gpglib.content_parsers.session_keys import PubSessionKeyParser

####################
### DELEGATORS
####################


class ContentParser(object):
    """Delegator to the different content parsers"""

    def __init__(self):
        self.parsers = {}
        self.parse_unknown = self.parser_for_unknown()

        # Instantiate parsers from find_parsers
        # It is recommended that this instance of ContentParser is memoized when created
        for tag_type, kls in self.find_parsers():
            self.parsers[tag_type] = kls()

    def parser_for_unknown(self):
        """Return instantiated parser to handle unknown tags"""
        return Parser()

    def find_parsers(self):
        """Specify lists of [(tag_type, kls), (tag_type, kls), ...] for kls to handle each tag type"""
        raise NotImplementedError()

    def consume(self, tag, message, kwargs):
        """
        Find parser given tag.tag_type
        And consume the body of the tag using correct packet parser
        """
        # Determine what parser to use for this packet
        # Default to self.parse_unknown, which will do some complaining for us
        parser = self.parsers.get(tag.tag_type, self.parse_unknown)

        # Consume the desired region
        return parser.consume(tag, message, tag.body, **kwargs)


class PacketContentParser(ContentParser):
    def find_parsers(self):
        """Specifiy parsers"""
        return (
            (1, PubSessionKeyParser),
            (2, SignatureParser),
            (5, SecretKeyParser),
            (6, PublicKeyParser),
            (7, SecretSubKeyParser),
            (8, CompressedParser),
            (9, SymEncryptedParser),
            (11, LiteralParser),
            (13, UserIdParser),
            (14, PublicSubKeyParser),
            (18, SymmetricallyEncryptedIntegrityProtectedDataPacketParser),
            (19, ModificationDetectionCodePacketParser),
        )


class SubSignatureContentParser(ContentParser):
    def find_parsers(self):
        """Don't handle any sub signature packets yet"""
        return ()

    def parser_for_unknown(self):
        """Return instantiated parser to handle unknown tags"""

        def consume(parser, tag, message, region, subsignature=None):
            if not subsignature:
                subsignature = {}
            subsignature[tag.tag_type] = region.read("bytes")
            return {"subsignature": subsignature}

        return type("SignatureParser", (Parser,), {"consume": consume})()



import bz2
import zlib

import bitstring
from Crypto import Random
from Crypto.Cipher import AES, CAST, DES3, Blowfish, PKCS1_v1_5
from Crypto.Hash import SHA, SHA256
from Crypto.PublicKey import DSA, RSA, ElGamal

from gpglib import errors
from gpglib.utils import bytes_to_long, long_to_bytes

####################
### CFB
####################


def crypt_CFB(region, ciphermod, key, iv):
    """
    Shamelessly stolen from OpenPGP (with some modifications)
    http://pypi.python.org/pypi/OpenPGP
    """
    # Create the cipher
    cipher = ciphermod.new(key, ciphermod.MODE_ECB)

    # Determine how many bytes to process at a time
    shift = ciphermod.block_size

    # Create a bitstring list of ['bytes:8', 'bytes:8', 'bytes:3']
    # Such that the entire remaining region length gets consumed
    region_length = (region.len - region.pos) // 8
    region_datas = ["bytes:%d" % shift] * (region_length // shift)
    leftover = region_length % shift
    if leftover:
        region_datas.append("bytes:%d" % (region_length % shift))

    # Use the cipher to decrypt region
    blocks = []
    for inblock in region.readlist(region_datas):
        mask = cipher.encrypt(iv)
        iv = inblock
        chunk = b"".join([bytes(bytearray((c ^ m,))) for m, c in zip(mask, inblock)])
        blocks.append(chunk)

    return b"".join(blocks)


####################
### MAPPINGS
####################


class Mapping(object):
    """
    Thin class that gives item access to some map of values
    That raises a NotImplementedError if you try to access something not defined on it
    """

    def __init__(self, typ, map):
        self.map = map
        self.type = typ

    def __getitem__(self, key):
        """Complain if key isn't known"""
        if key not in self.map:
            raise NotImplementedError("Haven't implemented %s : %s" % (self.type, key))
        return self.map[key]


class Algorithms(object):
    encryption = Mapping(
        "Symmetric encryption algorithm",
        {
            2: (DES3, 21),  # TripleDES 168 bit key derived from 192
            3: (CAST, 16),  # CAST5 128-bit key
            4: (Blowfish, 16),  # Blowfish 128-bit key
            7: (AES, 16),  # AES 128-bit key
            8: (AES, 24),  # AES with 192-bit key
            9: (AES, 32),  # AES with 256-bit key
        },
    )

    hashes = Mapping("Hash Algorithm", {2: SHA, 8: SHA256})  # SHA-1  # SHA-256

    keys = Mapping(
        "Key algorithm",
        {
            1: RSA,  # Encrypt or Sign
            2: RSA,  # Encrypt Only
            3: RSA,  # Sign Only
            16: ElGamal,  # Encrypt Only
            17: DSA,  # Digital Signature Algorithm
        },
    )


class Compression(object):
    def decompress_zip(compressed):
        """
        To decompress zip, we use zlib with a -15 window size.
        It says to ignore the zlib header
        and that the data is compressed with up to 15 bits of compression.
        """
        return zlib.decompress(compressed, -15)

    decompression = Mapping(
        "Decompressor", {1: decompress_zip, 2: zlib.decompress, 3: bz2.decompress}
    )


class Mapped(object):
    algorithms = Algorithms
    compression = Compression


####################
### PKCS
####################


class PKCS(object):
    @classmethod
    def consume(cls, region, key_algorithm, key):
        """
        Get next mpi values from region as according to key_algorithm
        Decrypt those mpis and then parse them as
        0x2 | random bytes | 0x0 | result

        The result will then be
        algorithm | session_key | checksum

        These values are retrieved from result and returned.
        If, however, mpis don't follow pattern above, then random bytes are used instead
        """
        # Get the mpi values from the region according to key_algorithm
        # And decrypt them with the provided key
        mpis = tuple(mpi.bytes for mpi in Mpi.consume_encryption(region, key_algorithm))

        if isinstance(key, RSA.RsaKey):
            sentinel = Random.new().read(255)
            val = mpis[0]
            if len(val) < 256:
                padding = b"\0" * (256 - len(val))
                val = padding + val
            bts = PKCS1_v1_5.new(key).decrypt(val, sentinel)
            decrypted = bitstring.ConstBitStream(bytes=bts)
        else:
            decrypted = cls.decrypt_elgamal(key, mpis)

        # The size of the key is the amount in decrypted
        # Minus the algorithm at the front and the checksum at the end
        key_size = (decrypted.len - decrypted.pos) / 8 - 1 - 2

        # The algorithm used to encrypt the message is the first byte
        # The session key is the next <key_size> bytes
        # The checksum is the last two bytes
        return decrypted.readlist("uint:8, bytes:%d, uint:16" "" % key_size)

    @classmethod
    def decrypt_elgamal(cls, key, mpis):
        """
        Elgamal is actually deprecated and so there isn't and won't be a
        public PKCS1_v1_5 type cypher for it

        So this method uses the _decrypt method on the key and manually unpads the result
        """
        mpis = list(map(bytes_to_long, mpis))
        bts = long_to_bytes(int(key._decrypt(mpis)))
        padded = bitstring.ConstBitStream(bytes=bts)

        # If decrypted isn't set by the end we raise an exception
        decrypted = None

        # First byte needs to be 02
        if padded.read("bytes:1") == b"\x02":
            # Find the next 00
            pos_before = padded.bytepos
            padded.find("0x00", bytealigned=True)
            pos_after = padded.bytepos

            # The ps section needs to be greater than 8
            if pos_after - pos_before >= 8:
                # Read in the seperator 0 byte
                # Gauranteed to be zero given use of find above
                padded.read("bytes:1")

                # Decrypted value is the rest of the padded value
                decrypted = padded

        if decrypted is None:
            # MPIs weren't valid, use random bytes instead
            decrypted = bitstring.ConstBitStream(bytes=Random.get_random_bytes(19))

        return decrypted


####################
### MPI VALUES
####################


class Mpi(object):
    """Object to hold logic for getting multi precision integers from a region"""

    @classmethod
    def parse(cls, region):
        """Retrieve one MPI value from the region"""
        # Get the length of the MPI to read in
        raw_mpi_length = region.read("uint:16")

        # Read in the MPI bytes and return the resulting bitstream
        mpi_length = (raw_mpi_length + 7) // 8
        return region.read(mpi_length * 8)

    @classmethod
    def retrieve(cls, region, mpis):
        """
        Helper to get multiple mpis from a region
        Allows some nice declarativity below....
        """
        return tuple(cls.parse(region) for mpi in mpis)

    ####################
    ### RFC4880 5.1
    ####################

    @classmethod
    def consume_encryption(cls, region, algorithm):
        """Retrieve necessary MPI values from a public session key"""
        if algorithm is RSA:
            return cls.retrieve(region, ("m**e mod n",))

        elif algorithm is ElGamal:
            return cls.retrieve(region, ("g**k mod p", "m * y**k mod p"))

        else:
            raise errors.PGPException("Unknown mpi algorithm for encryption %d" % algorithm)

    ####################
    ### RFC4880 5.5.2 and 5.5.3
    ####################

    @classmethod
    def consume_public(cls, region, algorithm):
        """Retrieve necessary MPI values from a public key for specified algorithm"""
        if algorithm is RSA:
            return cls.retrieve(region, ("n", "e"))

        elif algorithm is ElGamal:
            return cls.retrieve(region, ("p", "g", "y"))

        elif algorithm is DSA:
            p, q, g, y = cls.retrieve(region, ("p", "q", "g", "y"))

            # Pycryptodome has a weird construct order
            return y, g, p, q

        else:
            raise errors.PGPException("Unknown mpi algorithm for public keys %d" % algorithm)

    @classmethod
    def consume_private(cls, region, algorithm):
        """Retrieve necessary MPI values from a secret key for specified algorithm"""
        if algorithm is RSA:
            return cls.retrieve(region, ("d", "p", "q", "r"))

        elif algorithm is ElGamal:
            return cls.retrieve(region, ("x",))

        elif algorithm is DSA:
            return cls.retrieve(region, ("x",))

        else:
            raise errors.PGPException("Unknown mpi algorithm for secret keys %d" % algorithm)



from Crypto.Hash import SHA

from gpglib import errors
from gpglib.content_parsers.base import Parser
from gpglib.content_parsers.crypt import Mapped


class CompressedParser(Parser):
    """Parse compressed packets"""

    def consume(self, tag, message, region):
        # Get the compression algorithm used
        # And the rest of the message
        algo, rest = region.readlist("uint:8, bytes")

        # Get decompressor and use it to decompress the packet
        decompressor = Mapped.compression.decompression[algo]
        uncompressed = decompressor(rest)

        # Parse the inner packet and return it
        return message.consume(uncompressed)


class SymEncryptedParser(Parser):
    """Parse symmetrically encrypted data packet"""

    def consume(self, tag, message, region, algo, session_key):
        # Get the encryption algorithm used
        cipher, _ = Mapped.algorithms.encryption[algo]

        # Find out the length of the IV
        iv_len = cipher.block_size + 2

        # Read in the encrypted IV
        # The ciphertext is what's left in `region` after the iv
        iv, ciphertext = region.readlist("bytes:%d, bytes" % iv_len)

        # Build the cipher object from the session key and the IV
        decryptor = cipher.new(session_key, cipher.MODE_OPENPGP, iv)

        # Decrypt the ciphertext
        decrypted = decryptor.decrypt(ciphertext)

        # Parse the inner packet and return it
        return message.consume(decrypted)


class LiteralParser(Parser):
    """Extracts various information from the packet but only returns the plaintext"""

    def consume(self, tag, message, region):
        # Is it binary ('b'), text ('t') or utf-8 text ('u')
        # The length of the filename in bytes
        format, filename_length = region.readlist("bytes:1, uint:8")

        # Read in filename
        # Read in the date (can mean anything. ie. creation, modification, or 0)
        filename, date = region.readlist("bytes:%d, uint:32" % filename_length)

        # Add the literal data to the list of decrypted plaintext
        message.add_plaintext(region.read("bytes"))


class UserIdParser(Parser):
    """Parses type-13 packets, which contain information about the user"""

    def consume(self, tag, message, region):
        message.userid = region.read("bytes")


class SymmetricallyEncryptedIntegrityProtectedDataPacketParser(Parser):
    """Parses type-18 packets"""

    def consume(self, tag, message, region, algo, session_key):
        version = region.read("uint:8")
        if version != 1:
            raise errors.PGPException("Expected integrity packet to have version 1")

        encrypted_data = region.read("bytes")

        # Get a cipher object we can use to decrypt the key (and fail if we can't)
        cipherkls, key_size = Mapped.algorithms.encryption[algo]

        block_size = cipherkls.block_size
        segment_size = block_size * 8
        cipher = cipherkls.new(
            session_key, cipherkls.MODE_CFB, b"\x00" * block_size, segment_size=segment_size
        )

        decrypted_iv = cipher.decrypt(encrypted_data[:block_size])
        decrypted_data = bytearray()

        offset = block_size
        first_block = cipher.decrypt(encrypted_data[offset : offset + block_size])

        offset += block_size
        iv_check = first_block[:2]
        if iv_check != decrypted_iv[-2:]:
            raise errors.PGPException("There was a mismatch in the integrity check")

        decrypted_data.extend(first_block[2:])

        padding = block_size - (len(encrypted_data) - offset) % block_size
        encrypted_data += b"\x00" * padding

        decrypted_data.extend(cipher.decrypt(encrypted_data[offset:]))
        decrypted_data = decrypted_data[:-padding]

        # We run into trouble here with partial body lengths and the MDC.
        # Let's make pretend.
        mdc_data = bytearray()
        main_data = bytes(decrypted_data)
        if decrypted_data[-22:-20] == b"\xd3\x14":
            # It's there. Hold it back.
            mdc_data = bytes(decrypted_data[-22:])
            main_data = bytes(decrypted_data[:-22])

        message.consume(main_data)
        if mdc_data:
            message.consume(mdc_data)

        if not hasattr(message, "mdc"):
            raise errors.PGPException(
                "Integrity protected message is missing modification detection code."
            )

        hash_ = SHA.new(decrypted_iv)
        hash_.update(decrypted_iv[-2:])
        hash_.update(decrypted_data[:-20])

        if message.mdc != hash_.digest():
            raise errors.PGPException(
                "Integrity protected message does not match modification detection code."
            )


class ModificationDetectionCodePacketParser(Parser):
    """Parses type-17 packets"""

    def consume(self, tag, message, region):
        # Just contains a 20 octet sha1 hash
        message.add_mdc(region.read("bytes:20"))

        # Hack to make sure this packet is last
        return dict(mdc_must_be_last=True)



from gpglib import errors
from gpglib.content_parsers.base import Parser
from gpglib.content_parsers.crypt import PKCS, Mapped


class PubSessionKeyParser(Parser):
    """Parse public session key packet"""

    def consume(self, tag, message, region):
        # Version of the packet we're parsing
        # The id of the key used to encrypt the session key
        # The public key algorithm used to encrypt the session key
        vs = region.readlist("uint:8,  uint:64, uint:8")
        version, key_id, key_algo = vs

        # Get key algorithm
        key_algorithm = Mapped.algorithms.keys[key_algo]

        # Get the key which was used to encrypt the session key
        key = message.keys.get(key_id)
        if not key:
            typ = key_algorithm.__name__
            typ = typ[typ.rfind(".") + 1 :]
            raise errors.PGPException(
                "Data was encrypted with %s key '%d', which was't found" % (typ, key_id)
            )

        # Use PKCS to consume the region and decrypt it
        algo, session_key, checksum = PKCS.consume(region, key_algorithm, key)

        # Generate a checksum from the session_key (section 5.1 in the RFC).
        # This involves summing up all the bytes of the session key
        # and `mod`ing it by 65536.
        session_key_lst = [session_key[i : i + 1] for i in range(len(session_key))]

        generated_checksum = sum(ord(i) for i in session_key_lst) % 65536

        # Compare the checksums and throw an error if they don't match
        if checksum != generated_checksum:
            raise errors.PGPException(
                "The decrypted session key was invalid (checksums didn't match)"
            )

        # Pass the session key bytes and the algorithm used on to the next parser
        return dict(algo=algo, session_key=session_key)



import bitstring
from Crypto.Hash import SHA

from gpglib import errors
from gpglib.content_parsers.base import Parser
from gpglib.content_parsers.crypt import Mapped, Mpi, crypt_CFB

####################
### SIGNATURE
####################


class SignatureParser(Parser):
    """Signature packets describes a binding between some public key and some data"""

    def consume(self, tag, message, region):
        # Get values
        vs = region.readlist("uint:8,uint:8,uint:8,uint:8,uint:16")
        version, signature_type, public_key_algorithm, hash_algorithm, hashed_subpacket_length = vs

        # Complain if any values haven't been implemented yet
        self.only_implemented(version, (4,), "version four signature packets")
        self.only_implemented(signature_type, (0x13, 0x18), "UserId and Subkey binding signatures")

        # Get key algorithm
        # algorithm = Mapped.algorithms.keys[public_key_algorithm]

        # Get hasher
        # hasher = Mapped.algorithms.hashes[hash_algorithm]

        # Determine hashed data
        subsignature = region.read(hashed_subpacket_length * 8)

        ## hashed_subpacket_data
        message.consume_subsignature(subsignature)

        # Not cyrptographically protected by signature
        # Should only contain advisory information
        unhashed_subpacket_length = region.read("uint:16")
        unhashed_subpacket_body = region.read(unhashed_subpacket_length * 8)

        ## unhashed_subpacket_data
        message.consume_subsignature(unhashed_subpacket_body)

        # Left 16 bits of the signed hash value provided for a heuristic test for valid signatures
        ## left_of_signed_hash
        region.read(8 * 2)

        # Get the mpi value for the RSA hash
        # RSA signature value m**d mod n
        ## mdn
        Mpi.parse(region).read("uint")


####################
### BASE KEY PARSER
####################


class KeyParser(Parser):
    """
    Public and Secret keys are the same except Secret keys has some extra information
    SubKeys have the same format
    Hence the base takes care of all the common things to both formats and delegates the rest to the subclass
    """

    def consume(self, tag, message, region):
        info = self.consume_common(tag, message, region)

        # Get individual mpi_values and the raw bytes from the region for those values
        # The raw stream is required for the fingerprint data when determining key id
        mpi_values, raw_mpi_bytes = self.get_mpi_values(region, info["algorithm"])
        info["mpi_values"] = mpi_values
        info["raw_mpi_bytes"] = raw_mpi_bytes

        # Consume the rest of the Key
        self.consume_rest(tag, message, region, info)
        self.add_value(message, info)

    def get_mpi_values(self, region, algorithm):
        """
        * Determine position before
        * Get individual mpi values
        * Determine how much was read
        * Reset region to what it was before
        * Return byte stream for the amount read in the first pass
        """
        pos_before = region.pos
        mpi_values = Mpi.consume_public(region, algorithm)

        pos_after = region.pos
        region.pos = pos_before
        mpi_length = (pos_after - pos_before) / 8
        raw_mpi_bytes = region.read("bytes:%d" % mpi_length)

        return mpi_values, raw_mpi_bytes

    def consume_rest(self, tag, message, region, info):
        """Have common things to all keys in info"""
        pass

    def add_value(self, message, info):
        """Used to add information for this key to the message"""
        raise NotImplementedError

    def determine_key_id(self, info):
        """Calculate the key id"""
        fingerprint_data = b"".join(
            [
                chr(info["key_version"]).encode(),
                bitstring.Bits(uint=info["ctime"], length=4 * 8).bytes,
                chr(info["key_algo"]).encode(),
                info["raw_mpi_bytes"],
            ]
        )

        fingerprint_length = len(fingerprint_data)
        fingerprint_data = b"".join(
            [
                b"\x99",
                chr((0xFFFF & fingerprint_length) >> 8).encode(),
                chr(0xFF & fingerprint_length).encode(),
                fingerprint_data,
            ]
        )

        fingerprint = SHA.new(fingerprint_data).hexdigest().upper()[-16:]
        return int(fingerprint, 16)

    def consume_common(self, tag, message, region):
        """Common to all key types"""
        # Version of the public key
        # Creation time of the secret key
        # Public key algorithm used by this key
        vs = region.readlist("uint:8,uint:32,uint:8")
        public_key_version, ctime, key_algo = vs

        # Only version 4 packets are supported
        if public_key_version != 4:
            raise NotImplementedError(
                "Public key versions != 4 are not supported. Upgrade your PGP!"
            )

        # Get Key algorithm
        key_algorithm = Mapped.algorithms.keys[key_algo]

        return dict(
            tag=tag,
            key_version=public_key_version,
            ctime=ctime,
            algorithm=key_algorithm,
            key_algo=key_algo,
        )


####################
### PUBLIC KEY
####################


class PublicKeyParser(KeyParser):
    def add_value(self, message, info):
        message.add_key(info)

    def consume_rest(self, tag, message, region, info):
        values = [i.read("uint") for i in info["mpi_values"]]
        info["key"] = info["algorithm"].construct(values)
        info["key_id"] = self.determine_key_id(info)


####################
### SECRET KEY
####################


class SecretKeyParser(PublicKeyParser):
    def consume_rest(self, tag, message, region, info):
        """Already have public key things"""
        s2k_type = region.read("uint:8")
        mpis = self.get_mpis(s2k_type, message, region, info)

        # Get the rest of the mpi values and combine with those from public portion
        mpi_values = Mpi.consume_private(mpis, info["algorithm"])
        mpi_tuple = info["mpi_values"] + mpi_values

        # Record key and key_id
        values = [i.read("uint") for i in mpi_tuple]
        info["key"] = info["algorithm"].construct(values)
        info["key_id"] = self.determine_key_id(info)

    def get_mpis(self, s2k_type, message, region, info):
        """Get mpi values from region given specified string to key type"""
        self.only_implemented(s2k_type, (0, 254), "Unencrypted and s2k_type 254")

        # If string to key is:
        #   * 0 :: key is not encrypted.
        #   * 254 or 255 :: key is value of the string-to-key specifier
        #   * otherwise :: key is type of symmetric encryption algorithm used.
        if s2k_type == 0:
            # Unencrypted!
            mpis = region

        elif s2k_type == 254:
            # Get the symmetric encryption algorithm used
            encryption_algo = region.read(8).uint

            # Get a cipher object we can use to decrypt the key (and fail if we can't)
            cipher, key_size = Mapped.algorithms.encryption[encryption_algo]

            # This is the passphrase used to decrypt the secret key
            key_passphrase = self.parse_s2k(region, key_size, message.passphrase(message, info))

            # The IV is the next `block_size` bytes
            iv = region.read(cipher.block_size * 8).bytes

            # Use the hacky crypt_CFB func to decrypt the MPIs
            result = crypt_CFB(region, cipher, key_passphrase, iv)
            decrypted = bitstring.ConstBitStream(bytes=result)

            # The decrypted bytes are in the format of:
            #   MPIs || 20-octet SHA1 hash
            # Read in the MPIs portion of this
            mpis = decrypted.read(decrypted.len - (8 * 20))

            # Hash the mpi bytes
            generated_hash = SHA.new(mpis.bytes).digest()

            # Read in the 'real' hash
            real_hash = decrypted.read("bytes:20")

            if generated_hash != real_hash:
                raise errors.PGPException("Secret key hashes don't match. Check your passphrase")

        return mpis


####################
### SUB KEYS
####################


class PublicSubKeyParser(PublicKeyParser):
    """Same format as Public Key"""

    def add_value(self, message, info):
        message.add_sub_key(info)


class SecretSubKeyParser(SecretKeyParser):
    """Same format as Secret Key"""

    def add_value(self, message, info):
        message.add_sub_key(info)



import itertools
import math

from gpglib.content_parsers.crypt import Mapped
from gpglib.utils import binary_type


class Parser(object):
    """Base Parser class"""

    def consume(self, tag, message, region, **kwargs):
        """To be implemented by subclasses to do something with the body of the parser"""
        raise NotImplementedError("Don't know about tag type %d" % tag.tag_type)

    def only_implemented(self, received, implemented, message):
        """
        Useful to raise NotImplementedError when we get a value for something that hasn't been implemented
        i.e., self.only_implemented(received, (1, 2, 3), "values 1, 2 and 3")
        Would raise NotImplementedError saying "haven't implemented values 1, 2, and 3" if received is none of those values
        """
        if received not in implemented:
            raise NotImplementedError(
                "%s |:| Sorry, haven't implemented value %s. Have only implemented %s."
                % (self.name, received, message)
            )

    @property
    def name(self):
        return self.__class__.__name__

    def parse_s2k(self, region, key_size, passphrase=None):
        """Get and use string to key specifier for region"""
        # string-to-key specifier'
        # Hash algorithm used by the string-to-key value
        # The salt value used for the hash
        # Count to determine how much data gets hashed
        vs = region.readlist("uint:8,uint:8,bytes:8,uint:8")
        s2k_specifier, s2k_hash_algo, salt, raw_count = vs

        # Make sure passphrase is also a binary type
        # Because we add it to the salt
        if not isinstance(passphrase, binary_type):
            passphrase = passphrase.encode()

        self.only_implemented(s2k_specifier, (3,), "String to key type 3")

        # Get a hash object we can use
        hasher = Mapped.algorithms.hashes[s2k_hash_algo]

        # The 'count' is the length of the data that gets hashed
        count = (16 + (raw_count & 15)) << ((raw_count >> 4) + 6)

        # Combine the salt and passphrase for the repeating pattern
        combined = salt + passphrase

        # Infinite for loop
        result = []
        for i in itertools.count():
            # Initialize the message, which is at a minimum:
            #   some nulls || salt || passphrase
            # Fill the rest of the message (up to `count`) with the string `salt + passphrase`
            repeat_count = int(math.ceil((count - i) / float(len(combined))))
            message = (b"\x00" * i) + combined * repeat_count

            # Now hash the message
            hsh = hasher.new(message[:count]).digest()

            # Append the message to the result, until len(result) == count
            size = min(len(hsh), key_size - len(result))
            result.append(hsh[0:size])

            # Break if the result is large enough
            if sum(len(m) for m in result) >= key_size:
                break

        return b"".join(result)



from Crypto.Util.number import long_to_bytes

string_types = (str,)
binary_type = bytes
long_to_bytes = long_to_bytes


def bytes_to_long(s):
    return int.from_bytes(s, "big")


def dump(bytes):
    """Return a readable string for a string of bytes"""
    r = []
    for i in range(len(bytes)):
        r.append("%02x" % ord(bytes[i]))
    return " ".join(r)


class ValueTracker(object):
    """Track heirarchial relations of some items"""

    def __init__(self):
        self._items = {"items": []}
        self._current = None

    def consumed(self, *keys, **modifiers):
        """Return list of consumed items"""
        return list(self.items(self._items, keys, modifiers))

    def start_item(self, item):
        """Record the start of an item"""
        parent = self._items
        if self._current:
            parent = self._current

        next_item = {"items": [], "info": item, "parent": parent}
        self._current = next_item

        parent["items"].append(next_item)

    def end_item(self):
        """Record that an item was finished"""
        if self._current:
            self._current = self._current["parent"]

    def items(self, items, keys, modifiers):
        """
        Get a list from the heirarchy of recorded items
        [[item, children], [item, children], ...]

        Where children is of the same format but for the children of item
        """
        if items:
            for item in items["items"]:
                info = self.values_from(item["info"], keys, modifiers)
                yield info, list(self.items(item, keys, modifiers))

    def values_from(self, info, keys, modifiers):
        """
        Extract wanted information from info
        Useful for debugging purposes.
        """
        # Return info if we don't want to get anything from it
        if not keys and not modifiers:
            return info

        # We want to get things from info
        result = {}
        for key in list(keys) + modifiers.keys():
            if type(info) is dict:
                val = info[key]
            else:
                val = getattr(info, key)
            result[key] = val
        for key, modifier in modifiers.items():
            result[key] = modifier(info, key, val)

        if not modifiers and len(keys) == 1:
            key = keys[0]
            return result[key]

        if not keys and len(modifiers) == 1:
            key = modifiers.keys()[0]
            return result[key]

        return result



from collections import namedtuple

import bitstring

from gpglib.utils import ValueTracker, binary_type, string_types

# Information obtained from an OpenPGP header
Tag = namedtuple("Tag", ("version", "tag_type", "body"))

####################
### BASE MESSAGE
####################


class PGPMessage(object):
    """
    Class to hold details about a pgp message:
    Whether that be keys or encrypted data

    Has method for consuming the data using a PacketParser as message.consume
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.tags = ValueTracker()

    ####################
    ### CONSUMING
    ####################

    @property
    def packet_consumer(self):
        """Memoized PacketParser"""
        if not hasattr(self, "_packet_consumer"):
            from gpglib.packet_parser import PacketParser

            self._packet_consumer = PacketParser()
        return self._packet_consumer

    @property
    def subsignature_consumer(self):
        """Memoized SubSignatureParser"""
        if not hasattr(self, "_subsignature_consumer"):
            from gpglib.packet_parser import SubSignatureParser

            self._subsignature_consumer = SubSignatureParser()
        return self._subsignature_consumer

    def consume(self, region):
        """
        Consume a message.
        If a string is passed in as region, it is converted to a bitstream for you
        """
        if isinstance(region, string_types):
            region = bitstring.ConstBitStream(bytes=region.encode())
        elif isinstance(region, binary_type):
            region = bitstring.ConstBitStream(bytes=region)

        self.packet_consumer.consume(self, region)

    def consume_subsignature(self, region):
        """
        Consume subsignature packets
        If a string is passed in as region, it is converted to a bitstream for you
        """
        if isinstance(region, string_types):
            region = bitstring.ConstBitStream(bytes=region.encode())
        elif isinstance(region, binary_type):
            region = bitstring.ConstBitStream(bytes=region)

        self.subsignature_consumer.consume(self, region)

    ####################
    ### ADDING TAGS
    ####################

    def start_tag(self, tag):
        """Record start of a new tag"""
        self.tags.start_item(tag)

    def end_tag(self):
        """Record end of a new tag"""
        self.tags.end_item()


####################
### ENCRYPTED MESSAGE
####################


class EncryptedMessage(PGPMessage):
    def __init__(self, keys):
        self.keys = keys
        super(EncryptedMessage, self).__init__()

    def reset(self):
        super(EncryptedMessage, self).reset()
        self._plaintext = []

    def decrypt(self, region):
        """
        Consume the provided data
        And return the plaintext on the message
        """
        # Reset the plaintext
        self.reset()

        # Consume the stream
        self.consume(region)
        return self.plaintext

    @property
    def plaintext(self):
        """
        Concatenate all plaintext found in the message
        Requires decrypt to have already been called
        """
        return b"".join(self._plaintext)

    def add_plaintext(self, plaintext):
        """
        More plaintext was found
        I think it's possible to have multiple literalpackets in one pgp message.
        I could be wrong about that....
        """
        self._plaintext.append(plaintext)

    def add_mdc(self, mdc):
        self.mdc = mdc


####################
### KEY
####################


class Key(PGPMessage):
    def __init__(self, passphrase=None):
        self._passphrase = passphrase
        super(Key, self).__init__()

    def reset(self):
        super(Key, self).reset()
        self.keys = ValueTracker()

    @property
    def passphrase(self):
        """Return a function to get the passphrase"""
        if not callable(self._passphrase):
            _passphrase = self._passphrase

            def get_passphrase(message, info):
                return _passphrase

            self._passphrase = get_passphrase
        return self._passphrase

    def parse(self, region):
        self.reset()
        self.consume(region)
        return self

    def key_dict(self, keys=None):
        if keys is None:
            keys = self.keys.consumed()

        result = {}
        for key, subkeys in keys:
            result[key["key_id"]] = key["key"]
            result.update(self.key_dict(subkeys))
        return result

    ####################
    ### ADDING KEYS
    ####################

    def add_key(self, info):
        """Start a new public key"""
        self.keys.end_item()
        self.keys.start_item(info)

    def add_sub_key(self, info):
        """Add a sub public key"""
        self.keys.start_item(info)



from gpglib.content_parsers import PacketContentParser, SubSignatureContentParser
from gpglib.errors import PGPFormatException
from gpglib.structures import Tag

####################
### PARSER BASE
####################


class Parser(object):
    """Base for PacketContentParser and SubSignatureParser"""

    content_parser_kls = None

    def __init__(self):
        self.content_parser = self.make_content_parser()

    def make_content_parser(self):
        """Used by subclasses to determine content parser"""
        parser = self.content_parser_kls()
        parser.find_parsers()
        return parser

    def start_tag(self, tag, message):
        """Called when a tag is started"""
        pass

    def end_tag(self, tag, message):
        """Called when a tag is ended"""
        pass

    def next_tag(self, region):
        """Called to get the next tag from the region"""
        raise NotImplementedError()

    def consume(self, message, region):
        """
        Consume provided region of data
        Done by continually reading in packets untill none left
        Use next_tag to determine information about each packet
        Use content_parser to actually parse the packet
        """
        kwargs = {}
        while region.pos != region.len:
            tag = self.next_tag(region)

            # Pass the results from the previous parser call to the next one
            self.start_tag(tag, message)
            kwargs = self.content_parser.consume(tag, message, kwargs) or {}
            self.end_tag(tag, message)
        return message


####################
### NORMAL PACKETS
####################


class PacketParser(Parser):
    """
     RFC 4880 Section 4 says that a message is made up of many packets.
     Where a packet consists of a packet header followed by the packet body.

     The packet header is made up of::
         * The first 8 bytes, which is referred to as the tag
         * The length of the rest of the packet

     This implementation calls the entire packet header a tag
     and will create a structures.Tag to represent it

     consume will take in the Message structure and the next region of bytes for consideration
        For the first run, region should be the entire file
        For packets that contain other packets, region will be the bytes for just that packet

    message.bytes will always be the bytes for the entire message
    """

    content_parser_kls = PacketContentParser

    def start_tag(self, tag, message):
        """Called when a tag is started"""
        message.start_tag(tag)

    def end_tag(self, tag, message):
        """Called when a tag is ended"""
        message.end_tag()

    def next_tag(self, region):
        """Determine the version, tag_type and body_bit_length of the next packet"""
        # Tag information is held by the first 8 bytes
        tag = region.read(8)

        # Get left bit and version from first two bits
        left_bit, version = tag.readlist("uint:1, uint:1")

        # The left-most bit *MUST* be 1
        if not left_bit:
            raise PGPFormatException("The left-most bit of the tag ('%x') was not 1" % left_bit)

        # How the tag is parsed changes between the two versions
        if version == 1:
            # Read the tag type as the next 6 bits
            tag_type = tag.read("uint:6")
            return self.parse_new_tag(tag_type, region)
        else:
            return self.parse_old_tag(tag, region)

    def parse_new_tag(self, tag_type, region):
        """
        The length of the packet is then determined by the next group of bytes
        """
        # We peek at the next byte to determine what type of length to get
        length_type = region.peek("uint:8")
        body_length = self.determine_new_body_length(length_type, region)

        # Determine the body of the packet
        if body_length is not None:
            body = region.read(body_length * 8)
        else:
            # Found a partial packet. Add up all the partials to get the entire body
            body_len = 1 << (length_type & 0b11111)

            # Read the specified length of bytes from the body
            body = region.read(body_len * 8)

            # See recursion
            body += self.parse_new_tag(tag_type, region).body

        # Return the tag
        return Tag(version=1, tag_type=tag_type, body=body)

    def parse_old_tag(self, tag, region):
        """
        6 bits left to parse in the tag
        Type is the first four
        and length is determined by the two after that
        """
        tag_type, length_type = tag.readlist("uint:4, uint:2")

        if length_type == 3:
            # indeterminate length untill the end of the file
            body_length = None
        else:
            # Determine the length of the packet body
            body_length = self.determine_old_body_length(length_type, region)

        # Get body of the packet
        if body_length is not None:
            body = region.read(body_length * 8)
        else:
            body = region.read(region.len - region.pos)

        # Return the tag
        return Tag(version=0, tag_type=tag_type, body=body)

    def determine_old_body_length(self, length_type, region):
        """Determine body length of an old style packet"""
        if length_type < 3:
            octet_length = 2**length_type
            return region.read("uint:%d" % (8 * octet_length))
        else:
            # indeterminate length untill the end of the file
            return None

    def determine_new_body_length(self, length_type, region):
        """
        The first byte (given as length_type and still to be read from region) is used to determine how many to look at
        < 192 = one octet
        > 192 and < 224  = two octet
        == 255 = ignore the 255, and use the next 4 octets
        otherwise it is partial length
        """
        if length_type < 192:
            return region.read("uint:8")

        elif length_type < 224:
            one, two = region.readlist("uint:8, uint:8")
            return ((one - 192) << 8) + (two + 192)

        elif length_type == 255:
            # Ignore the first octet (255 just says to look at next 4)
            region.read(8)

            # Add up the next four octets
            return region.read("uint:32")

        else:
            # Length_type hasn't been read yet, just peeked
            region.read(8)

            # Return None to specify a partial packet
            return None


####################
### SUB SIGNATURE PACKETS
####################


class SubSignatureParser(Parser):
    """SubSignaturePackets are a bit different to normal packets"""

    content_parser_kls = SubSignatureContentParser

    def next_tag(self, region):
        """Called to get the next tag from the region"""
        # First few octets says the length of the subpacket
        body_length = self.determine_body_length(region)

        # After length is the tag type and then the body of the message
        body = region.read(body_length * 8)
        tag_type = body.read("uint:8")

        return Tag(version=None, tag_type=tag_type, body=body)

    def determine_body_length(self, region):
        """RFC 4880 5.2.3.1"""
        length_type = region.peek("uint:8")

        if length_type < 192:
            return region.read("uint:8")

        elif length_type < 255:
            one, two = region.readlist("uint:8, uint:8")
            return ((one - 192) << 8) + (two + 192)

        elif length_type == 255:
            # Ignore the first octet (255 just says to look at next 4)
            region.read(8)

            # Add up the next four octets
            return region.read("uint:32")
